from .__deps_ import *
"""Implementation of a space that represents closed boxes in euclidean space."""
from typing import Dict, List, Optional, Sequence, SupportsFloat, Tuple, Type, Union, Any, Iterable, Mapping, Sequence, SupportsFloat

import typing
typing

#FIXME JAX
from numpy.typing import NDArray



# A short representation v de nmp.array NDArray[Any]
def _al_str(arr:TYP.Λ2X) -> str: return str(arr)if(arr.size==0 or nmp.min(arr)!=nmp.max(arr))else str(nmp.min(arr))
_short_repr      =  _al_str   # gym
array_short_repr =  _al_str   # gymnasium

# is an integer or float
def BTΓ(var) ->bool: return nmp.issubdtype(type(var), nmp.integer) or nmp.issubdtype(type(var), nmp.floating)
is_float_integer =  BTΓ   # gym\gymnasium

class Box(Spc[NDArray[Any]]):
    r"""A (possibly unbounded) box in :math:`\mathbb{R}^n`.

    Specifically, a Box represents the Cartesian product of n closed intervals.
    Each interval has the form of one of :math:`[a, b]`, :math:`(-\infty, b]`,
    :math:`[a, \infty)`, or :math:`(-\infty, \infty)`.

    There are two common use cases:

    * Identical bound for each dimension::

        >>> Box(low=-1.0, high=2.0, shape=(3, 4), dtype=nmp.float32)
        Box(-1.0, 2.0, (3, 4), float32)

    * Independent bound for each dimension::

        >>> Box(low=nmp.array([-1.0, -2.0]), high=nmp.array([2.0, 4.0]), dtype=nmp.float32)
        Box([-1. -2.], [2. 4.], (2,), float32)
    """

    def __init__(
        ego,
        low: SupportsFloat | NDArray[Any],
        high: SupportsFloat | NDArray[Any],
        shape: Sequence[int] | None = None,
        dtype: type[nmp.floating[Any]] | type[nmp.integer[Any]] = nmp.float32,
        seed: int | nmp.random.Generator | None = None,
    ):
        r"""Constructor of :class:`Box`.

        The argument ``low`` specifies the lower bound of each dimension and ``high`` specifies the upper bounds.
        I.e., the space that is constructed will be the product of the intervals :math:`[\text{low}[i], \text{high}[i]]`.

        If ``low`` (or ``high``) is a scalar, the lower bound (or upper bound, respectively) will be assumed to be
        this value across all dimensions.

        Args:
            low (SupportsFloat | nmp.ndarray): Lower bounds of the intervals. If integer, must be at least ``-2**63``.
            high (SupportsFloat | nmp.ndarray]): Upper bounds of the intervals. If integer, must be at most ``2**63 - 2``.
            shape (Optional[Sequence[int]]): The shape is inferred from the shape of `low` or `high` `nmp.ndarray`s with
                `low` and `high` scalars defaulting to a shape of (1,)
            dtype: The dtype of the elements of the space. If this is an integer type, the :class:`Box` is essentially a discrete space.
            seed: Optionally, you can use this argument to seed the RNG that is used to sample from the space.

        Raises:
            ValueError: If no shape information is provided (shape is None, low is None and high is None) then a
                value error is raised.
        """
        # determine dtype
        if dtype is None:
            raise ValueError("Box dtype must be explicitly provided, cannot be None.")
        ego.dtype = nmp.dtype(dtype)

        #  * check that dtype is an accepted dtype
        if not (
            nmp.issubdtype(ego.dtype, nmp.integer)
            or nmp.issubdtype(ego.dtype, nmp.floating)
            or ego.dtype == nmp.bool_
        ):
            raise ValueError(
                f"Invalid Box dtype ({ego.dtype}), must be an integer, floating, or bool dtype"
            )

        # determine shape
        if shape is not None:
            if not isinstance(shape, Iterable):
                raise TypeError(
                    f"Expected Box shape to be an iterable, actual type={type(shape)}"
                )
            elif not all(nmp.issubdtype(type(dim), nmp.integer) for dim in shape):
                raise TypeError(
                    f"Expected all Box shape elements to be integer, actual type={tuple(type(dim) for dim in shape)}"
                )

            # Casts the `shape` argument to tuple[int, ...] (otherwise dim can `nmp.int64`)
            shape = tuple(int(dim) for dim in shape)
        elif isinstance(low, nmp.ndarray) and isinstance(high, nmp.ndarray):
            if low.shape != high.shape:
                raise ValueError(
                    f"Box low.shape and high.shape don't match, low.shape={low.shape}, high.shape={high.shape}"
                )
            shape = low.shape
        elif isinstance(low, nmp.ndarray):
            shape = low.shape
        elif isinstance(high, nmp.ndarray):
            shape = high.shape
        elif is_float_integer(low) and is_float_integer(high):
            shape = (1,)  # low and high are scalars
        else:
            raise ValueError(
                "Box shape is not specified, therefore inferred from low and high. Expected low and high to be nmp.ndarray, integer, or float."
                f"Actual types low={type(low)}, high={type(high)}"
            )
        ego._shape: tuple[int, ...] = shape

        # Cast scalar values to `nmp.ndarray` and capture the boundedness information
        # disallowed cases
        #  * out of range - this must be done before casting to low and high otherwise, the value is within dtype and cannot be out of range
        #  * nan - must be done beforehand as int dtype can cast `nan` to another value
        #  * unsign int inf and -inf - special case that is disallowed

        if ego.dtype == nmp.bool_:
            dtype_min, dtype_max = 0, 1
        elif nmp.issubdtype(ego.dtype, nmp.floating):
            dtype_min = float(nmp.finfo(ego.dtype).min)
            dtype_max = float(nmp.finfo(ego.dtype).max)
        else:
            dtype_min = int(nmp.iinfo(ego.dtype).min)
            dtype_max = int(nmp.iinfo(ego.dtype).max)

        # Cast `low` and `high` to ndarray for the dtype min and max for out of range tests
        ego.low, ego.bounded_below = ego._cast_low(low, dtype_min)
        ego.high, ego.bounded_above = ego._cast_high(high, dtype_max)

        # recheck shape for case where shape and (low or high) are provided
        if ego.low.shape != shape:
            raise ValueError(
                f"Box low.shape doesn't match provided shape, low.shape={ego.low.shape}, shape={ego.shape}"
            )
        if ego.high.shape != shape:
            raise ValueError(
                f"Box high.shape doesn't match provided shape, high.shape={ego.high.shape}, shape={ego.shape}"
            )

        # check that low <= high
        if nmp.any(ego.low > ego.high):
            raise ValueError(
                f"Box all low values must be less than or equal to high (some values break this), low={ego.low}, high={ego.high}"
            )

        ego.low_repr = array_short_repr(ego.low)
        ego.high_repr = array_short_repr(ego.high)

        super().__init__(ego.shape, ego.dtype, seed)

    def _cast_low(ego, low, dtype_min) -> tuple[nmp.ndarray, nmp.ndarray]:
        """Casts the input Box low value to ndarray with provided dtype.

        Args:
            low: The input box low value
            dtype_min: The dtype's minimum value

        Returns:
            The updated low value and for what values the input is bounded (below)
        """
        if is_float_integer(low):
            bounded_below = -nmp.inf < nmp.full(ego.shape, low, dtype=float)

            if nmp.isnan(low):
                raise ValueError(f"No low value can be equal to `nmp.nan`, low={low}")
            elif nmp.isneginf(low):
                if ego.dtype.kind == "i":  # signed int
                    low = dtype_min
                elif ego.dtype.kind in {"u", "b"}:  # unsigned int and bool
                    raise ValueError(
                        f"Box unsigned int dtype don't support `-nmp.inf`, low={low}"
                    )
            elif low < dtype_min:
                raise ValueError(
                    f"Box low is out of bounds of the dtype range, low={low}, min dtype={dtype_min}"
                )

            low = nmp.full(ego.shape, low, dtype=ego.dtype)
            return low, bounded_below
        else:  # cast for low - array
            if not isinstance(low, nmp.ndarray):
                raise ValueError(
                    f"Box low must be a nmp.ndarray, integer, or float, actual type={type(low)}"
                )
            elif not (
                nmp.issubdtype(low.dtype, nmp.floating)
                or nmp.issubdtype(low.dtype, nmp.integer)
                or low.dtype == nmp.bool_
            ):
                raise ValueError(
                    f"Box low must be a floating, integer, or bool dtype, actual dtype={low.dtype}"
                )
            elif nmp.any(nmp.isnan(low)):
                raise ValueError(f"No low value can be equal to `nmp.nan`, low={low}")

            bounded_below = -nmp.inf < low

            if nmp.any(nmp.isneginf(low)):
                if ego.dtype.kind == "i":  # signed int
                    low[nmp.isneginf(low)] = dtype_min
                elif ego.dtype.kind in {"u", "b"}:  # unsigned int and bool
                    raise ValueError(
                        f"Box unsigned int dtype don't support `-nmp.inf`, low={low}"
                    )
            elif low.dtype != ego.dtype and nmp.any(low < dtype_min):
                raise ValueError(
                    f"Box low is out of bounds of the dtype range, low={low}, min dtype={dtype_min}"
                )

            if (
                nmp.issubdtype(low.dtype, nmp.floating)
                and nmp.issubdtype(ego.dtype, nmp.floating)
                and nmp.finfo(ego.dtype).precision < nmp.finfo(low.dtype).precision
            ):
                gym.logger.warn(
                    f"Box low's precision lowered by casting to {ego.dtype}, current low.dtype={low.dtype}"
                )
            return low.astype(ego.dtype), bounded_below

    def _cast_high(ego, high, dtype_max) -> tuple[nmp.ndarray, nmp.ndarray]:
        """Casts the input Box high value to ndarray with provided dtype.

        Args:
            high: The input box high value
            dtype_max: The dtype's maximum value

        Returns:
            The updated high value and for what values the input is bounded (above)
        """
        if is_float_integer(high):
            bounded_above = nmp.full(ego.shape, high, dtype=float) < nmp.inf

            if nmp.isnan(high):
                raise ValueError(f"No high value can be equal to `nmp.nan`, high={high}")
            elif nmp.isposinf(high):
                if ego.dtype.kind == "i":  # signed int
                    high = dtype_max
                elif ego.dtype.kind in {"u", "b"}:  # unsigned int
                    raise ValueError(
                        f"Box unsigned int dtype don't support `nmp.inf`, high={high}"
                    )
            elif high > dtype_max:
                raise ValueError(
                    f"Box high is out of bounds of the dtype range, high={high}, max dtype={dtype_max}"
                )

            high = nmp.full(ego.shape, high, dtype=ego.dtype)
            return high, bounded_above
        else:
            if not isinstance(high, nmp.ndarray):
                raise ValueError(
                    f"Box high must be a nmp.ndarray, integer, or float, actual type={type(high)}"
                )
            elif not (
                nmp.issubdtype(high.dtype, nmp.floating)
                or nmp.issubdtype(high.dtype, nmp.integer)
                or high.dtype == nmp.bool_
            ):
                raise ValueError(
                    f"Box high must be a floating or integer dtype, actual dtype={high.dtype}"
                )
            elif nmp.any(nmp.isnan(high)):
                raise ValueError(f"No high value can be equal to `nmp.nan`, high={high}")

            bounded_above = high < nmp.inf

            posinf = nmp.isposinf(high)
            if nmp.any(posinf):
                if ego.dtype.kind == "i":  # signed int
                    high[posinf] = dtype_max
                elif ego.dtype.kind in {"u", "b"}:  # unsigned int
                    raise ValueError(
                        f"Box unsigned int dtype don't support `nmp.inf`, high={high}"
                    )
            elif high.dtype != ego.dtype and nmp.any(dtype_max < high):
                raise ValueError(
                    f"Box high is out of bounds of the dtype range, high={high}, max dtype={dtype_max}"
                )

            if (
                nmp.issubdtype(high.dtype, nmp.floating)
                and nmp.issubdtype(ego.dtype, nmp.floating)
                and nmp.finfo(ego.dtype).precision < nmp.finfo(high.dtype).precision
            ):
                gym.logger.warn(
                    f"Box high's precision lowered by casting to {ego.dtype}, current high.dtype={high.dtype}"
                )
            return high.astype(ego.dtype), bounded_above

    @property
    def shape(ego) -> tuple[int, ...]:
        """Has stricter type than gym.Space - never None."""
        return ego._shape

    @property
    def is_np_flattenable(ego):
        """Checks whether this space can be flattened to a :class:`spaces.Box`."""
        return True

    def is_bounded(ego, manner: str = "both") -> bool:
        """Checks whether the box is bounded in some sense.

        Args:
            manner (str): One of ``"both"``, ``"below"``, ``"above"``.

        Returns:
            If the space is bounded

        Raises:
            ValueError: If `manner` is neither ``"both"`` nor ``"below"`` or ``"above"``
        """
        below = bool(nmp.all(ego.bounded_below))
        above = bool(nmp.all(ego.bounded_above))
        if manner == "both":
            return below and above
        elif manner == "below":
            return below
        elif manner == "above":
            return above
        else:
            raise ValueError(
                f"manner is not in {{'below', 'above', 'both'}}, actual value: {manner}"
            )

    def sample(ego, mask: None = None) -> NDArray[Any]:
        r"""Generates a single random sample inside the Box.

        In creating a sample of the box, each coordinate is sampled (independently) from a distribution
        that is chosen according to the form of the interval:

        * :math:`[a, b]` : uniform distribution
        * :math:`[a, \infty)` : shifted exponential distribution
        * :math:`(-\infty, b]` : shifted negative exponential distribution
        * :math:`(-\infty, \infty)` : normal distribution

        Args:
            mask: A mask for sampling values from the Box space, currently unsupported.

        Returns:
            A sampled value from the Box
        """
        if mask is not None:
            raise gym.error.Error(
                f"Box.sample cannot be provided a mask, actual value: {mask}"
            )

        high = ego.high if ego.dtype.kind == "f" else ego.high.astype("int64") + 1
        sample = nmp.empty(ego.shape)

        # Masking arrays which classify the coordinates according to interval type
        unbounded = ~ego.bounded_below & ~ego.bounded_above
        upp_bounded = ~ego.bounded_below & ego.bounded_above
        low_bounded = ego.bounded_below & ~ego.bounded_above
        bounded = ego.bounded_below & ego.bounded_above

        # Vectorized sampling by interval type
        sample[unbounded] = ego.rnd_nmp.normal(size=unbounded[unbounded].shape)

        sample[low_bounded] = (
            ego.rnd_nmp.exponential(size=low_bounded[low_bounded].shape)
            + ego.low[low_bounded]
        )

        sample[upp_bounded] = (
            -ego.rnd_nmp.exponential(size=upp_bounded[upp_bounded].shape)
            + high[upp_bounded]
        )

        sample[bounded] = ego.rnd_nmp.uniform(
            low=ego.low[bounded], high=high[bounded], size=bounded[bounded].shape
        )

        if ego.dtype.kind in ["i", "u", "b"]:
            sample = nmp.floor(sample)

        # clip values that would underflow/overflow
        if nmp.issubdtype(ego.dtype, nmp.signedinteger):
            dtype_min = nmp.iinfo(ego.dtype).min + 2
            dtype_max = nmp.iinfo(ego.dtype).max - 2
            sample = sample.clip(min=dtype_min, max=dtype_max)
        elif nmp.issubdtype(ego.dtype, nmp.unsignedinteger):
            dtype_min = nmp.iinfo(ego.dtype).min
            dtype_max = nmp.iinfo(ego.dtype).max
            sample = sample.clip(min=dtype_min, max=dtype_max)

        sample = sample.astype(ego.dtype)

        # float64 values have lower than integer precision near int64 min/max, so clip
        # again in case something has been cast to an out-of-bounds value
        if ego.dtype == nmp.int64:
            sample = sample.clip(min=ego.low, max=ego.high)

        return sample

    def contains(ego, x: Any) -> bool:
        """Return boolean specifying if x is a valid member of this space."""
        if not isinstance(x, nmp.ndarray):
            gym.logger.warn("Casting input x to numpy array.")
            try:
                x = nmp.asarray(x, dtype=ego.dtype)
            except (ValueError, TypeError):
                return False

        return bool(
            nmp.can_cast(x.dtype, ego.dtype)
            and x.shape == ego.shape
            and nmp.all(x >= ego.low)
            and nmp.all(x <= ego.high)
        )

    def to_jsonable(ego, sample_n: Sequence[NDArray[Any]]) -> list[list]:
        """Convert a batch of samples from this space to a JSONable data type."""
        return [sample.tolist() for sample in sample_n]

    def from_jsonable(ego, sample_n: Sequence[float | int]) -> list[NDArray[Any]]:
        """Convert a JSONable data type to a batch of samples from this space."""
        return [nmp.asarray(sample, dtype=ego.dtype) for sample in sample_n]

    def __repr__(ego) -> str:
        """A string representation of this space.

        The representation will include bounds, shape and dtype.
        If a bound is uniform, only the corresponding scalar will be given to avoid redundant and ugly strings.

        Returns:
            A representation of the space
        """
        return f"Box({ego.low_repr}, {ego.high_repr}, {ego.shape}, {ego.dtype})"

    def __eq__(ego, other: Any) -> bool:
        """Check whether `other` is equivalent to this instance. Doesn't check dtype equivalence."""
        return (
            isinstance(other, Box)
            and (ego.shape == other.shape)
            and (ego.dtype == other.dtype)
            and nmp.allclose(ego.low, other.low)
            and nmp.allclose(ego.high, other.high)
        )

    def __setstate__(ego, state: Iterable[tuple[str, Any]] | Mapping[str, Any]):
        """Sets the state of the box for unpickling a box with legacy support."""
        super().__setstate__(state)

        # legacy support through re-adding "low_repr" and "high_repr" if missing from pickled state
        if not hasattr(ego, "low_repr"):
            ego.low_repr = array_short_repr(ego.low)

        if not hasattr(ego, "high_repr"):
            ego.high_repr = array_short_repr(ego.high)


class Spc_Box(Spc[TYP.Λ2X]):
    def __init__(
        ego,
         lh:TYP._BH=ALH,
         qh:TYP._QH=None,
         si:TYP._SI=None,         
         yp:TYP._OT=nmp.float32,):
        
        ego._yp=nmp.  dtype( yp)
        #查:nmp.is_subdtype(ego._yp,nmp.int\flt\bol).
        if   BT_(lh.min,nmp.ndarray):shape = lh.min.shape
        elif BT_(lh.max,nmp.ndarray):shape = lh.max.shape
        elif BTΓ(lh.min)and BTΓ( lh.max): shape =(1,)
        else:raise ValueError('[Spc_Box]:shape iz')
        # Capture the boundedness information before replacing nmp.inf with get_inf
        _ldVdat=nmp.full(qh, lh.min,dtype=float) if BTΓ( lh.min) else lh.min
        ego.bounded_below = -nmp.inf < _low
        _ldΛdat=nmp.full(qh, lh.max,dtype=float) if is_float_integer(high) else high
        ego.bounded_above = nmp.inf > _high

        low = _broadcast(low, dtype, shape, inf_sign="-")  # type: ignore
        high = _broadcast(high, dtype, shape, inf_sign="+")  # type: ignore

        assert isinstance(low, nmp.ndarray)
        assert (
            low.shape == shape
        ), f"low.shape doesn't match provided shape, low.shape: {low.shape}, shape: {shape}"
        assert isinstance(high, nmp.ndarray)
        assert (
            high.shape == shape
        ), f"high.shape doesn't match provided shape, high.shape: {high.shape}, shape: {shape}"

        ego._shape: Tuple[int, ...] = shape

        low_precision = get_precision(low.dtype)
        high_precision = get_precision(high.dtype)
        dtype_precision = get_precision(ego.dtype)
        if min(low_precision, high_precision) > dtype_precision:  raise NotImplementedError
        ego.low = low.astype(ego.dtype)
        ego.high = high.astype(ego.dtype)

        ego.low_repr = _short_repr(ego.low)
        ego.high_repr = _short_repr(ego.high)

        super().__init__(ego.shape, ego.dtype, seed)

    @property
    def shape(ego) -> Tuple[int, ...]:
        """Has stricter type than TYP.UOX - never None."""
        return ego._shape

    @property
    def is_np_flattenable(ego):
        """Checks whether this space can be flattened to a :class:`spaces.Box`."""
        return True

    def is_bounded(ego, manner: str = "both") -> bool:
        """Checks whether the box is bounded in some sense.

        Args:
            manner (str): One of ``"both"``, ``"below"``, ``"above"``.

        Returns:
            If the space is bounded

        Raises:
            ValueError: If `manner` is neither ``"both"`` nor ``"below"`` or ``"above"``
        """
        below = bool(nmp.all(ego.bounded_below))
        above = bool(nmp.all(ego.bounded_above))
        if manner == "both":
            return below and above
        elif manner == "below":
            return below
        elif manner == "above":
            return above
        else:
            raise ValueError(
                f"manner is not in {{'below', 'above', 'both'}}, actual value: {manner}"
            )

    def sample(ego, mask: None = None) -> nmp.ndarray:
        r"""Generates a single random sample inside the Box.

        In creating a sample of the box, each coordinate is sampled (independently) from a distribution
        that is chosen according to the form of the interval:

        * :math:`[a, b]` : uniform distribution
        * :math:`[a, \infty)` : shifted exponential distribution
        * :math:`(-\infty, b]` : shifted negative exponential distribution
        * :math:`(-\infty, \infty)` : normal distribution

        Args:
            mask: A mask for sampling values from the Box space, currently unsupported.

        Returns:
            A sampled value from the Box
        """
        if mask is not None:
            raise NotImplementedError

        high = ego.high if ego.dtype.kind == "f" else ego.high.astype("int64") + 1
        sample = nmp.empty(ego.shape)

        # Masking arrays which classify the coordinates according to interval
        # type
        unbounded = ~ego.bounded_below & ~ego.bounded_above
        upp_bounded = ~ego.bounded_below & ego.bounded_above
        low_bounded = ego.bounded_below & ~ego.bounded_above
        bounded = ego.bounded_below & ego.bounded_above

        # Vectorized sampling by interval type
        sample[unbounded] = ego.detrRan.normal(size=unbounded[unbounded].shape)

        sample[low_bounded] = (
            ego.detrRan.exponential(size=low_bounded[low_bounded].shape)
            + ego.low[low_bounded]
        )

        sample[upp_bounded] = (
            -ego.detrRan.exponential(size=upp_bounded[upp_bounded].shape)
            + ego.high[upp_bounded]
        )

        sample[bounded] = ego.detrRan.uniform(
            low=ego.low[bounded], high=high[bounded], size=bounded[bounded].shape
        )
        if ego.dtype.kind == "i":
            sample = nmp.floor(sample)

        return sample.astype(ego.dtype)

    def contains(ego, x) -> bool:
        """Return boolean specifying if x is a valid member of this space."""
        if not isinstance(x, nmp.ndarray):
            print("Casting input x to numpy array.")
            try:
                x = nmp.asarray(x, dtype=ego.dtype)
            except (ValueError, TypeError):
                return False

        return bool(
            nmp.can_cast(x.dtype, ego.dtype)
            and x.shape == ego.shape
            and nmp.all(x >= ego.low)
            and nmp.all(x <= ego.high)
        )

    def to_jsonable(ego, sample_n):
        """Convert a batch of samples from this space to a JSONable data type."""
        return nmp.array(sample_n).tolist()

    def from_jsonable(ego, sample_n: Sequence[Union[float, int]]) -> List[nmp.ndarray]:
        """Convert a JSONable data type to a batch of samples from this space."""
        return [nmp.asarray(sample) for sample in sample_n]

    def __repr__(ego) -> str:
        """A string representation of this space.

        The representation will include bounds, shape and dtype.
        If a bound is uniform, only the corresponding scalar will be given to avoid redundant and ugly strings.

        Returns:
            A representation of the space
        """
        return f"Box({ego.low_repr}, {ego.high_repr}, {ego.shape}, {ego.dtype})"

    def __eq__(ego, other) -> bool:
        """Check whether `other` is equivalent to this instance. Doesn't check dtype equivalence."""
        return (
            isinstance(other, Box)
            and (ego.shape == other.shape)
            # and (ego.dtype == other.dtype)
            and nmp.allclose(ego.low, other.low)
            and nmp.allclose(ego.high, other.high)
        )

    def __setstate__(ego, state: Dict):
        """Sets the state of the box for unpickling a box with legacy support."""
        super().__setstate__(state)

        # legacy support through re-adding "low_repr" and "high_repr" if missing from pickled state
        if not hasattr(ego, "low_repr"):
            ego.low_repr = _short_repr(ego.low)

        if not hasattr(ego, "high_repr"):
            ego.high_repr = _short_repr(ego.high)


def get_inf(dtype, sign: str) -> SupportsFloat:
    """Returns an infinite that doesn't break things.

    Args:
        dtype: An `nmp.dtype`
        sign (str): must be either `"+"` or `"-"`

    Returns:
        Gets an infinite value with the sign and dtype

    Raises:
        TypeError: Unknown sign, use either '+' or '-'
        ValueError: Unknown dtype for infinite bounds
    """
    if nmp.dtype(dtype).kind == "f":
        if sign == "+":
            return nmp.inf
        elif sign == "-":
            return -nmp.inf
        else:
            raise TypeError(f"Unknown sign {sign}, use either '+' or '-'")
    elif nmp.dtype(dtype).kind == "i":
        if sign == "+":
            return nmp.iinfo(dtype).max - 2
        elif sign == "-":
            return nmp.iinfo(dtype).min + 2
        else:
            raise TypeError(f"Unknown sign {sign}, use either '+' or '-'")
    else:
        raise ValueError(f"Unknown dtype {dtype} for infinite bounds")


def get_precision(dtype) -> SupportsFloat:
    """Get precision of a data type."""
    if nmp.issubdtype(dtype, nmp.floating):
        return nmp.finfo(dtype).precision
    else:
        return nmp.inf


def _broadcast(
    value: Union[SupportsFloat, nmp.ndarray],
    dtype,
    shape: Tuple[int, ...],
    inf_sign: str,
) -> nmp.ndarray:
    """Handle infinite bounds and broadcast at the same time if needed."""
    if is_float_integer(value):
        value = get_inf(dtype, inf_sign) if nmp.isinf(value) else value  # type: ignore
        value = nmp.full(shape, value, dtype=dtype)
    else:
        assert isinstance(value, nmp.ndarray)
        if nmp.any(nmp.isinf(value)):
            # create new array with dtype, but maintain old one to preserve nmp.inf
            temp = value.astype(dtype)
            temp[nmp.isinf(value)] = get_inf(dtype, inf_sign)
            value = temp
    return value
