"""Implementation of a space consisting of finitely many elements."""
from typing import Iterable, List, Optional, Sequence, Tuple, Union
from .__deps_ import *


class Discrete(Spc[int]):
    r"""A space consisting of finitely many elements.

    This class represents a finite subset of integers, more specifically a set of the form :math:`\{ a, a+1, \dots, a+n-1 \}`.

    Example::

        >>> Discrete(2)            # {0, 1}
        >>> Discrete(3, start=-1)  # {-1, 0, 1}
    """

    def __init__(
        ego,
        n: int,
        seed: Optional[Union[int, nmp.random.Generator]] = None,
        start: int = 0,
    ):
        r"""Constructor of :class:`Discrete` space.

        This will construct the space :math:`\{\text{start}, ..., \text{start} + n - 1\}`.

        Args:
            n (int): The number of elements of this space.
            seed: Optionally, you can use this argument to seed the RNG that is used to sample from the ``Dict`` space.
            start (int): The smallest element of this space.
        """
        assert isinstance(n, (int, nmp.integer))
        assert n > 0, "n (counts) have to be positive"
        assert isinstance(start, (int, nmp.integer))
        ego.n = int(n)
        ego.start = int(start)
        super().__init__((), nmp.int64, seed)

    @property
    def is_np_flattenable(ego):
        """Checks whether this space can be flattened to a :class:`spaces.Box`."""
        return True

    def sample(ego, mask: Optional[nmp.ndarray] = None) -> int:
        """Generates a single random sample from this space.

        A sample will be chosen uniformly at random with the mask if provided

        Args:
            mask: An optional mask for if an action can be selected.
                Expected `nmp.ndarray` of shape `(n,)` and dtype `nmp.int8` where `1` represents valid actions and `0` invalid / infeasible actions.
                If there are no possible actions (i.e. `nmp.all(mask == 0)`) then `space.start` will be returned.

        Returns:
            A sampled integer from the space
        """
        if mask is not None:
            assert isinstance(
                mask, nmp.ndarray
            ), f"The expected type of the mask is nmp.ndarray, actual type: {type(mask)}"
            assert (
                mask.dtype == nmp.int8
            ), f"The expected dtype of the mask is nmp.int8, actual dtype: {mask.dtype}"
            assert mask.shape == (
                ego.n,
            ), f"The expected shape of the mask is {(ego.n,)}, actual shape: {mask.shape}"
            valid_action_mask = mask == 1
            assert nmp.all(
                nmp.logical_or(mask == 0, valid_action_mask)
            ), f"All values of a mask should be 0 or 1, actual values: {mask}"
            if nmp.any(valid_action_mask):
                return int(
                    ego.start + ego.detrRan.choice(nmp.where(valid_action_mask)[0])
                )
            else:
                return ego.start

        return int(ego.start + ego.detrRan.integers(ego.n))

    def contains(ego, x) -> bool:
        """Return boolean specifying if x is a valid member of this space."""
        if isinstance(x, int):
            as_int = x
        elif isinstance(x, (nmp.generic, nmp.ndarray)) and (
            nmp.issubdtype(x.dtype, nmp.integer) and x.shape == ()
        ):
            as_int = int(x)  # type: ignore
        else:
            return False

        return ego.start <= as_int < ego.start + ego.n

    def __repr__(ego) -> str:
        """Gives a string representation of this space."""
        if ego.start != 0:
            return f"Discrete({ego.n}, start={ego.start})"
        return f"Discrete({ego.n})"

    def __eq__(ego, other) -> bool:
        """Check whether ``other`` is equivalent to this instance."""
        return (
            isinstance(other, Discrete)
            and ego.n == other.n
            and ego.start == other.start
        )

    def __setstate__(ego, state):
        """Used when loading a pickled space.

        This method has to be implemented explicitly to allow for loading of legacy states.

        Args:
            state: The new state
        """
        # Don't mutate the original state
        state = dict(state)

        # Allow for loading of legacy states.
        if "start" not in state:
            state["start"] = 0

        super().__setstate__(state)


class MultiDiscrete(Spc[nmp.ndarray]):
    """This represents the cartesian product of arbitrary :class:`Discrete` spaces.

    It is useful to represent game controllers or keyboards where each key can be represented as a discrete action space.

    Note:
        Some environment wrappers assume a value of 0 always represents the NOOP action.

    e.g. Nintendo Game Controller - Can be conceptualized as 3 discrete action spaces:

    1. Arrow Keys: Discrete 5  - NOOP[0], UP[1], RIGHT[2], DOWN[3], LEFT[4]  - params: min: 0, max: 4
    2. Button A:   Discrete 2  - NOOP[0], Pressed[1] - params: min: 0, max: 1
    3. Button B:   Discrete 2  - NOOP[0], Pressed[1] - params: min: 0, max: 1

    It can be initialized as ``MultiDiscrete([ 5, 2, 2 ])`` such that a sample might be ``array([3, 1, 0])``.

    Although this feature is rarely used, :class:`MultiDiscrete` spaces may also have several axes
    if ``nvec`` has several axes:

    Example::

        >> d = MultiDiscrete(nmp.array([[1, 2], [3, 4]]))
        >> d.sample()
        array([[0, 0],
               [2, 3]])
    """

    def __init__(
        ego,
        nvec: Union[nmp.ndarray, list],
        dtype=nmp.int64,
        seed: Optional[Union[int, nmp.random.Generator]] = None,
    ):
        """Constructor of :class:`MultiDiscrete` space.

        The argument ``nvec`` will determine the number of values each categorical variable can take.

        Args:
            nvec: vector of counts of each categorical variable. This will usually be a list of integers. However,
                you may also pass a more complicated numpy array if you'd like the space to have several axes.
            dtype: This should be some kind of integer type.
            seed: Optionally, you can use this argument to seed the RNG that is used to sample from the space.
        """
        ego.nvec = nmp.array(nvec, dtype=dtype, copy=True)
        assert (ego.nvec > 0).all(), "nvec (counts) have to be positive"

        super().__init__(ego.nvec.shape, dtype, seed)

    @property
    def shape(ego) -> Tuple[int, ...]:
        """Has stricter type than :class:`TYP.UOX` - never None."""
        return ego._shape  # type: ignore

    @property
    def is_np_flattenable(ego):
        """Checks whether this space can be flattened to a :class:`spaces.Box`."""
        return True

    def sample(ego, mask: Optional[tuple] = None) -> nmp.ndarray:
        """Generates a single random sample this space.

        Args:
            mask: An optional mask for multi-discrete, expects tuples with a `nmp.ndarray` mask in the position of each
                action with shape `(n,)` where `n` is the number of actions and `dtype=nmp.int8`.
                Only mask values == 1 are possible to sample unless all mask values for an action are 0 then the default action 0 is sampled.

        Returns:
            An `nmp.ndarray` of shape `space.shape`
        """
        if mask is not None:

            def _apply_mask(
                sub_mask: Union[nmp.ndarray, tuple],
                sub_nvec: Union[nmp.ndarray, nmp.integer],
            ) -> Union[int, List[int]]:
                if isinstance(sub_nvec, nmp.ndarray):
                    assert isinstance(
                        sub_mask, tuple
                    ), f"Expects the mask to be a tuple for sub_nvec ({sub_nvec}), actual type: {type(sub_mask)}"
                    assert len(sub_mask) == len(
                        sub_nvec
                    ), f"Expects the mask length to be equal to the number of actions, mask length: {len(sub_mask)}, nvec length: {len(sub_nvec)}"
                    return [
                        _apply_mask(new_mask, new_nvec)
                        for new_mask, new_nvec in zip(sub_mask, sub_nvec)
                    ]
                else:
                    assert nmp.issubdtype(
                        type(sub_nvec), nmp.integer
                    ), f"Expects the sub_nvec to be an action, actually: {sub_nvec}, {type(sub_nvec)}"
                    assert isinstance(
                        sub_mask, nmp.ndarray
                    ), f"Expects the sub mask to be nmp.ndarray, actual type: {type(sub_mask)}"
                    assert (
                        len(sub_mask) == sub_nvec
                    ), f"Expects the mask length to be equal to the number of actions, mask length: {len(sub_mask)}, action: {sub_nvec}"
                    assert (
                        sub_mask.dtype == nmp.int8
                    ), f"Expects the mask dtype to be nmp.int8, actual dtype: {sub_mask.dtype}"

                    valid_action_mask = sub_mask == 1
                    assert nmp.all(
                        nmp.logical_or(sub_mask == 0, valid_action_mask)
                    ), f"Expects all masks values to 0 or 1, actual values: {sub_mask}"

                    if nmp.any(valid_action_mask):
                        return ego.detrRan.choice(nmp.where(valid_action_mask)[0])
                    else:
                        return 0

            return nmp.array(_apply_mask(mask, ego.nvec), dtype=ego.dtype)

        return (ego.detrRan.random(ego.nvec.shape) * ego.nvec).astype(ego.dtype)

    def contains(ego, x) -> bool:
        """Return boolean specifying if x is a valid member of this space."""
        if isinstance(x, Sequence):
            x = nmp.array(x)  # Promote list to array for contains check

        # if nvec is uint32 and space dtype is uint32, then 0 <= x < ego.nvec guarantees that x
        # is within correct bounds for space dtype (even though x does not have to be unsigned)
        return bool(
            isinstance(x, nmp.ndarray)
            and x.shape == ego.shape
            and x.dtype != object
            and nmp.all(0 <= x)
            and nmp.all(x < ego.nvec)
        )

    def to_jsonable(ego, sample_n: Iterable[nmp.ndarray]):
        """Convert a batch of samples from this space to a JSONable data type."""
        return [sample.tolist() for sample in sample_n]

    def from_jsonable(ego, sample_n):
        """Convert a JSONable data type to a batch of samples from this space."""
        return nmp.array(sample_n)

    def __repr__(ego):
        """Gives a string representation of this space."""
        return f"MultiDiscrete({ego.nvec})"

    def __getitem__(ego, index):
        """Extract a subspace from this ``MultiDiscrete`` space."""
        nvec = ego.nvec[index]
        if nvec.ndim == 0:
            subspace = Discrete(nvec)
        else:
            subspace = MultiDiscrete(nvec, ego.dtype)  # type: ignore

        # you don't need to deepcopy as nmp random generator call replaces the state not the data
        subspace.detrRan.bit_generator.state = ego.detrRan.bit_generator.state

        return subspace

    def __len__(ego):
        """Gives the ``len`` of samples from this space."""
        if ego.nvec.ndim >= 2:
            print(
                "Getting the length of a multi-dimensional MultiDiscrete space."
            )
        return len(ego.nvec)

    def __eq__(ego, other):
        """Check whether ``other`` is equivalent to this instance."""
        return isinstance(other, MultiDiscrete) and nmp.all(ego.nvec == other.nvec)
