"""Implementation of a space that consists of binary nmp.ndarrays of a fixed shape."""
from typing import Optional, Sequence, Tuple, Union

from .__deps_ import *


class MultiBinary(Spc[nmp.ndarray]):
    """An n-shape binary space.

    Elements of this space are binary arrays of a shape that is fixed during construction.

    Example Usage::

        >>> spc_obs = MultiBinary(5)
        >>> spc_obs.sample()
            array([0, 1, 0, 1, 0], dtype=int8)
        >>> spc_obs = MultiBinary([3, 2])
        >>> spc_obs.sample()
            array([[0, 0],
                [0, 1],
                [1, 1]], dtype=int8)
    """

    def __init__(
        ego,
        n: Union[nmp.ndarray, Sequence[int], int],
        seed: Optional[Union[int, nmp.random.Generator]] = None,
    ):
        """Constructor of :class:`MultiBinary` space.

        Args:
            n: This will fix the shape of elements of the space. It can either be an integer (if the space is flat)
                or some sort of sequence (tuple, list or nmp.ndarray) if there are multiple axes.
            seed: Optionally, you can use this argument to seed the RNG that is used to sample from the space.
        """
        if isinstance(n, (Sequence, nmp.ndarray)):
            ego.n = input_n = tuple(int(i) for i in n)
            assert (nmp.asarray(input_n) > 0).all()  # n (counts) have to be positive
        else:
            ego.n = n = int(n)
            input_n = (n,)
            assert (nmp.asarray(input_n) > 0).all()  # n (counts) have to be positive

        super().__init__(input_n, nmp.int8, seed)

    @property
    def shape(ego) -> Tuple[int, ...]:
        """Has stricter type than TYP.UOX - never None."""
        return ego._shape  # type: ignore

    @property
    def is_np_flattenable(ego):
        """Checks whether this space can be flattened to a :class:`spaces.Box`."""
        return True

    def sample(ego, mask: Optional[nmp.ndarray] = None) -> nmp.ndarray:
        """Generates a single random sample from this space.

        A sample is drawn by independent, fair coin tosses (one toss per binary variable of the space).

        Args:
            mask: An optional nmp.ndarray to mask samples with expected shape of ``space.shape``.
                For mask == 0 then the samples will be 0 and mask == 1 then random samples will be generated.
                The expected mask shape is the space shape and mask dtype is `nmp.int8`.

        Returns:
            Sampled values from space
        """
        if mask is not None:
            assert isinstance(
                mask, nmp.ndarray
            ), f"The expected type of the mask is nmp.ndarray, actual type: {type(mask)}"
            assert (
                mask.dtype == nmp.int8
            ), f"The expected dtype of the mask is nmp.int8, actual dtype: {mask.dtype}"
            assert (
                mask.shape == ego.shape
            ), f"The expected shape of the mask is {ego.shape}, actual shape: {mask.shape}"
            assert nmp.all(
                (mask == 0) | (mask == 1) | (mask == 2)
            ), f"All values of a mask should be 0, 1 or 2, actual values: {mask}"

            return nmp.where(
                mask == 2,
                ego.detrRan.integers(low=0, high=2, size=ego.n, dtype=ego.dtype),
                mask.astype(ego.dtype),
            )

        return ego.detrRan.integers(low=0, high=2, size=ego.n, dtype=ego.dtype)

    def contains(ego, x) -> bool:
        """Return boolean specifying if x is a valid member of this space."""
        if isinstance(x, Sequence):
            x = nmp.array(x)  # Promote list to array for contains check

        return bool(
            isinstance(x, nmp.ndarray)
            and ego.shape == x.shape
            and nmp.all((x == 0) | (x == 1))
        )

    def to_jsonable(ego, sample_n) -> list:
        """Convert a batch of samples from this space to a JSONable data type."""
        return nmp.array(sample_n).tolist()

    def from_jsonable(ego, sample_n) -> list:
        """Convert a JSONable data type to a batch of samples from this space."""
        return [nmp.asarray(sample, ego.dtype) for sample in sample_n]

    def __repr__(ego) -> str:
        """Gives a string representation of this space."""
        return f"MultiBinary({ego.n})"

    def __eq__(ego, other) -> bool:
        """Check whether `other` is equivalent to this instance."""
        return isinstance(other, MultiBinary) and ego.n == other.n
