"""Implementation of a space that represents the cartesian product of other spaces."""
from collections import namedtuple
from collections.abc import Sequence as CollectionSequence
from typing import Iterable, Optional
from typing import Sequence as TypingSequence
from typing import Tuple as TypingTuple
from typing import Union

import numpy as nmp

from .__deps_ import *

class Spc_NyH(Spc[tuple],CollectionSequence):
    
    def __init__(go, em:Iterable[Spc],yh:tuple=None,
        seed:Optional[Union[int, TypingSequence[int], nmp.random.Generator]] = None):
        if isinstance(em,dict):yh,em=zip(*em.items())
        go._yh=yh or tuple(['l_'+str(_)for _ in DIT(em)])
        go._jc ='NyTuple'
        globals()[go._jc]=namedtuple(go._jc,go._yh)
        go._yp=globals()[go._jc]
        #go._em=em
        go._em=go._yp(*em)
        for space in go._em:assert isinstance(space, Spc),\
            "Elements of the tuple must be instances of TYP.UOX"
        super().__init__(None, None, seed)  # type: ignore

    @property
    def spaces(ego):
        return ego._em

    @property
    def is_np_flattenable(ego):
        """Checks whether this space can be flattened to a :class:`spaces.Box`."""
        return all(space.is_np_flattenable for space in ego._em)
    @property
    def tqh_shp(ego):
        return {ego._yh[_]: tuple(ego._em[_].shape) for _ in DIT(ego)}

    def seed(
        ego, seed: Optional[Union[int, TypingSequence[int]]] = None
    ) -> TypingSequence[int]:
        """Seed the PRNG of this space and all subspaces.

        Depending on the type of seed, the subspaces will be seeded differently
        * None - All the subspaces will use a random initial seed
        * Int - The integer is used to seed the `Tuple` space that is used to generate seed values for each of the subspaces. Warning, this does not guarantee unique seeds for all of the subspaces.
        * List - Values used to seed the subspaces. This allows the seeding of multiple composite subspaces (`List(42, 54, ...)`).

        Args:
            seed: An optional list of ints or int to seed the (sub-)spaces.
        """
        seeds = []

        if isinstance(seed, CollectionSequence):
            assert len(seed) == len(
                ego._em
            ), f"Expects that the subspaces of seeds equals the number of subspaces. Actual length of seeds: {len(seeds)}, length of subspaces: {len(ego._em)}"
            for subseed, space in zip(seed, ego._em):
                seeds += space.seed(subseed)
        elif isinstance(seed, int):
            seeds = super().seed(seed)
            subseeds = ego.detrRan.integers(
                nmp.iinfo(nmp.int32).max, size=len(ego._em)
            )
            for subspace, subseed in zip(ego._em, subseeds):
                seeds += subspace.seed(int(subseed))
        elif seed is None:
            for space in ego._em:
                seeds += space.seed(seed)
        else:
            raise TypeError(
                f"Expected seed type: list, tuple, int or None, actual type: {type(seed)}"
            )

        return seeds

    def sample(
        ego, mask: Optional[TypingTuple[Optional[nmp.ndarray], ...]] = None
    ) -> tuple:
        """Sample a random element from this space."""
        #return type(ego)(zip(self.names, super().sample()))
        if mask is not None:
            assert isinstance(
                mask, tuple
            ), f"Expected type of mask is tuple, actual type: {type(mask)}"
            assert len(mask) == len(
                ego._em
            ), f"Expected length of mask is {len(ego._em)}, actual length: {len(mask)}"

            return ego._yp(
                space.sample(mask=sub_mask)
                for space, sub_mask in zip(ego._em, mask)
            )

        return ego._yp(space.sample() for space in ego._em)

    def contains(ego, x) -> bool:
        """Return boolean specifying if x is a valid member of this space."""
        if isinstance(x, (list, nmp.ndarray)):
            x = tuple(x)  # Promote list and ndarray to tuple for contains check
        return (
            isinstance(x, tuple)
            and len(x) == len(ego._em)
            and all(space.contains(part) for (space, part) in zip(ego._em, x))
        )

    # TODO 抽象
    def __repr__(ego) -> str:
        """Gives a string representation of this space."""
        return f"{ego._jc}(" + ", ".join([str(s) for s in ego._em]) + ")"
    def to_jsonable(ego, sample_n: CollectionSequence) -> list:
        """Convert a batch of samples from this space to a JSONable data type."""
        # serialize as list-repr of tuple of vectors
        return [
            space.to_jsonable([sample[i] for sample in sample_n])
            for i, space in enumerate(ego._em)
        ]

    def from_jsonable(ego, sample_n) -> list:
        """Convert a JSONable data type to a batch of samples from this space."""
        return [
            sample
            for sample in zip(
                *[
                    space.from_jsonable(sample_n[i])
                    for i, space in enumerate(ego._em)
                ]
            )
        ]

    def __getitem__(ego, lx) -> Spc:
        """Get the subspace at specific `lx`."""
        if isinstance(lx,int):return ego._em[lx]
        return ego._em[ego._yh.index(lx)]
    
    def __getattr__(ego, lx) -> Spc:
        """Get the subspace at specific `lx`."""
        if isinstance(lx,int):return ego._em[lx]
        return ego._em[ego._yh.index(lx)]
    # Get the number of subspaces (?cartesian product)
    def __len__(ego):return len(ego._em)
    
    def __eq__( ego, od) -> bool:
        """Check whether ``other`` is equivalent to this instance."""
        return(isinstance(od,Spc_NyH)and ego._em==od._em)or\
              (isinstance(od,Tuple)and ego._em==od.spaces)or\
              (isinstance(od,tuple)and ego._em==od)


class Tuple(Spc[tuple], CollectionSequence):
    """A tuple (more precisely: the cartesian product) of :class:`Spc` instances.

    Elements of this space are tuples of elements of the constituent spaces.

    Example usage::

        >>> from num.spaces import Box, Discrete
        >>> spc_obs = Tuple((Discrete(2), Box(-1, 1, shape=(2,))))
        >>> spc_obs.sample()
        (0, array([0.03633198, 0.42370757], dtype=float32))
    """

    def __init__(
        ego,
        spaces: Iterable[Spc],
        seed: Optional[Union[int, TypingSequence[int], nmp.random.Generator]] = None,
    ):
        r"""Constructor of :class:`Tuple` space.

        The generated instance will represent the cartesian product :math:`\text{spaces}[0] \times ... \times \text{spaces}[-1]`.

        Args:
            spaces (Iterable[Spc]): The spaces that are involved in the cartesian product.
            seed: Optionally, you can use this argument to seed the RNGs of the ``spaces`` to ensure reproducible sampling.
        """
        ego.spaces = tuple(spaces)
        for space in ego.spaces:
            assert isinstance(
                space, Spc
            ), "Elements of the tuple must be instances of TYP.UOX"
        super().__init__(None, None, seed)  # type: ignore

    @property
    def is_np_flattenable(ego):
        """Checks whether this space can be flattened to a :class:`spaces.Box`."""
        return all(space.is_np_flattenable for space in ego.spaces)

    def seed(
        ego, seed: Optional[Union[int, TypingSequence[int]]] = None
    ) -> TypingSequence[int]:
        """Seed the PRNG of this space and all subspaces.

        Depending on the type of seed, the subspaces will be seeded differently
        * None - All the subspaces will use a random initial seed
        * Int - The integer is used to seed the `Tuple` space that is used to generate seed values for each of the subspaces. Warning, this does not guarantee unique seeds for all of the subspaces.
        * List - Values used to seed the subspaces. This allows the seeding of multiple composite subspaces (`List(42, 54, ...)`).

        Args:
            seed: An optional list of ints or int to seed the (sub-)spaces.
        """
        seeds = []

        if isinstance(seed, CollectionSequence):
            assert len(seed) == len(
                ego.spaces
            ), f"Expects that the subspaces of seeds equals the number of subspaces. Actual length of seeds: {len(seeds)}, length of subspaces: {len(ego.spaces)}"
            for subseed, space in zip(seed, ego.spaces):
                seeds += space.seed(subseed)
        elif isinstance(seed, int):
            seeds = super().seed(seed)
            subseeds = ego.detrRan.integers(
                nmp.iinfo(nmp.int32).max, size=len(ego.spaces)
            )
            for subspace, subseed in zip(ego.spaces, subseeds):
                seeds += subspace.seed(int(subseed))
        elif seed is None:
            for space in ego.spaces:
                seeds += space.seed(seed)
        else:
            raise TypeError(
                f"Expected seed type: list, tuple, int or None, actual type: {type(seed)}"
            )

        return seeds

    def sample(
        ego, mask: Optional[TypingTuple[Optional[nmp.ndarray], ...]] = None
    ) -> tuple:
        """Generates a single random sample inside this space.

        This method draws independent samples from the subspaces.

        Args:
            mask: An optional tuple of optional masks for each of the subspace's samples,
                expects the same number of masks as spaces

        Returns:
            Tuple of the subspace's samples
        """
        if mask is not None:
            assert isinstance(
                mask, tuple
            ), f"Expected type of mask is tuple, actual type: {type(mask)}"
            assert len(mask) == len(
                ego.spaces
            ), f"Expected length of mask is {len(ego.spaces)}, actual length: {len(mask)}"

            return tuple(
                space.sample(mask=sub_mask)
                for space, sub_mask in zip(ego.spaces, mask)
            )

        return tuple(space.sample() for space in ego.spaces)

    def contains(ego, x) -> bool:
        """Return boolean specifying if x is a valid member of this space."""
        if isinstance(x, (list, nmp.ndarray)):
            x = tuple(x)  # Promote list and ndarray to tuple for contains check
        return (
            isinstance(x, tuple)
            and len(x) == len(ego.spaces)
            and all(space.contains(part) for (space, part) in zip(ego.spaces, x))
        )

    def __repr__(ego) -> str:
        """Gives a string representation of this space."""
        return "Tuple(" + ", ".join([str(s) for s in ego.spaces]) + ")"

    def to_jsonable(ego, sample_n: CollectionSequence) -> list:
        """Convert a batch of samples from this space to a JSONable data type."""
        # serialize as list-repr of tuple of vectors
        return [
            space.to_jsonable([sample[i] for sample in sample_n])
            for i, space in enumerate(ego.spaces)
        ]

    def from_jsonable(ego, sample_n) -> list:
        """Convert a JSONable data type to a batch of samples from this space."""
        return [
            sample
            for sample in zip(
                *[
                    space.from_jsonable(sample_n[i])
                    for i, space in enumerate(ego.spaces)
                ]
            )
        ]

    def __getitem__(ego, index: int) -> Spc:
        """Get the subspace at specific `index`."""
        return ego.spaces[index]

    def __len__(ego) -> int:
        """Get the number of subspaces that are involved in the cartesian product."""
        return len(ego.spaces)

    def __eq__(ego, other) -> bool:
        """Check whether ``other`` is equivalent to this instance."""
        return isinstance(other, Tuple) and ego.spaces == other.spaces
